#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define MAXTOKEN 100             /* can't go below 3 */
#define MAXOUTPUT 1000           /* should always be >= MAXTOKEN */

enum { NAME, PARENS, BRACKETS } TokenType;
enum { OK, ERROR };

static const char *Typelist[] = {
    "char",
    "int",
    "short",
    "long",
    "unsigned",
    "float",
    "double",
    "void",
    NULL
};

static char Name[MAXTOKEN], Output[MAXOUTPUT], Token[MAXTOKEN];
static int OutSpc = MAXOUTPUT;

void undeclaration(void);
void declaration(void);

int dcl(void);
int dirdcl(void);

void outputCat(const char *s);
void flushline(void);
int gettoken(void);

int main(const int argc, const char *argv[])
{
    if (argc > 1 && strcmp(argv[1], "--undcl") == 0)
        undeclaration();
    else
        declaration();
    return 0;
}

void undeclaration(void)
{
    char type, temp[MAXOUTPUT];
    int ispointer = 0;

    while (gettoken() != EOF) {
        strcpy(Output, Token);
        OutSpc = MAXOUTPUT - strlen(Token);
        while ((type = gettoken()) != '\n') {
            if (type == PARENS || type == BRACKETS) {
                if (ispointer) {
                    OutSpc -= snprintf(temp, MAXOUTPUT, "(%s)%s", Output, Token); /* C99 feature */
                    strcpy(Output, temp);
                    ispointer = 0;
                } else {
                    outputCat(Token);
                }
            } else if (type == '*') {
                snprintf(temp, MAXOUTPUT, "*%s", Output);
                strcpy(Output, temp);
                ispointer = 1;
            } else if (type == NAME) {
                snprintf(temp, MAXOUTPUT, "%s %s", Token, Output);
                strcpy(Output, temp);
                ispointer = 0;
            } else {
                printf("ERROR: Invalid input at %s\n", Token);
            }
        }
        printf("%s\n", Output);
        ispointer = 0;          /* reset for next line of input */
    }
}

char gettype(char *s);

void declaration(void)
{
    char datatype[MAXTOKEN];

    while (gettype(datatype) != EOF) {
        Output[0] = '\0';
        if (dcl() == ERROR) {
            flushline();
        } else if (TokenType != '\n' && TokenType != ',') {
            printf("ERROR: Syntax error for %s.\n", datatype);
            flushline();
        } else {
            printf("%s: %s %s\n", Name, Output, datatype);
            Token[0] = '\0';
        }
    }
}

int dcl(void)
{
    int ns = 0;

    while (gettoken() == '*')
        ++ns;
    if (dirdcl() == ERROR)
        return ERROR;
    while (ns-- > 0)
        outputCat(" pointer to");

    return OK;
}

int dirdcl(void)
{
    if (TokenType == '(') {
        if (dcl() == ERROR) {
            return ERROR;
        }
        if (TokenType != ')') {
            printf("ERROR: Closing ')' paren missing.\n");
            return ERROR;
        }
    } else if (TokenType == NAME) {
        strcpy(Name, Token);
    } else {
        printf("ERROR: Expected name or (dcl).\n");
        return ERROR;
    }

    while (gettoken() == PARENS || TokenType == BRACKETS) {
        if (TokenType == PARENS) {
            outputCat(" function returning");
        } else {
            outputCat(" array");
            strcat(Output, Token);
            outputCat(" of");
        }
    }

    return OK;
}

char getch(void);
void ungetch(const char c);

int gettoken(void)
{
    char c, *p = Token;

    while (isblank(c = getch())) /* C99 feature */
        ;
    if (c == '(') {
        if ((c = getch()) == ')') {
            strcpy(Token, "()");
            TokenType = PARENS;
        } else {
            ungetch(c);
            TokenType = '(';
        }
    } else if (c == '[') {
        for (*p++ = c; (*p++ = getch()) != ']' && p-Token < MAXTOKEN-1; )
            ;
        *p = '\0';
        TokenType = BRACKETS;
    } else if (isalpha(c) || c == '_') {
        for (*p++ = c; (isalnum(c = getch()) || c == '_') && p-Token < MAXTOKEN-1; ) {
            *p++ = c;
        }
        *p = '\0';
        ungetch(c);
        TokenType = NAME;
    } else {
        TokenType = c;
    }

    return TokenType;
}

char gettype(char *s)
{
    static const char *last, *start, **type;
    static char c;

    while (isblank(c = getch()))
        ;

    last = start = s;
    while (isalpha(c) || c == '_') {
        last = s;
        for (*s++ = c; (isalnum(c = getch()) || c == '_') && s-start < MAXTOKEN-1; ) {
            *s++ = c;
        }
        if (isblank(c) && s-start < MAXTOKEN-1) {
            *s++ = ' ';     /* no point in printing a tab char */
            while (isblank(c = getch()))
                ;
        }
    }
    ungetch(c);
    if (*(s-1) == ' ')      /* should only be one 'space' at most */
        --s;
    *s = '\0';

    if (last != start) {
        for (type = Typelist; *type != NULL && strcmp(*type, last); ++type)
            ;
        if (*type == NULL) {
            do {
                ungetch(*--s);
            } while (s >= last);
            *s = '\0';
        }
    }

    TokenType = '\n';
    return c;
}

void outputCat(const char *s)
{
    if ((OutSpc -= strlen(s)) > 0) {
        strcat(Output, s);
    } else {
        printf("ERROR: Not enough space in Output[] to add:\n\
\"%s\"\n", s);
    }
}

void flushline(void)
{
    static char c;

    /* flush both buffer[] and stdin */
    while ((c = getch()) != '\n' && c != EOF)
        ;
    if (c == EOF)
        ungetch(c);
}

#define BUFFER 1000

char buffer[BUFFER];
char *bp = buffer;

char getch(void)
{
    return (bp > buffer) ? *--bp : getchar();
}

void ungetch(const char c)
{
    if (bp - buffer < BUFFER)
        *bp++ = c;
    else
        printf("ERROR: ungetch(); not enough room to unget '%c'.\n");
}
